module Syntax where

infix 10 _≡ˠ_
infix 10 _≡⁺_
infix 10 _≡ˢ_
infix 10 _≡_
infixl 50 _•_

mutual 
  data Con : Set where
    ε   : Con
    _,_ : forall Γ -> Ty Γ -> Con

  data _≡ˠ_ : Con -> Con -> Set where
    congε : ε ≡ˠ ε
    cong, : forall {Γ Γ' σ σ'} -> Γ ≡ˠ Γ' -> σ ≡⁺ σ' -> (Γ , σ) ≡ˠ (Γ' , σ')

  data Ty : Con -> Set where
    U     : forall {Γ} -> Ty Γ
    El    : forall {Γ} -> Tm Γ U -> Ty Γ
    Π     : forall {Γ} σ -> Ty (Γ , σ) -> Ty Γ

  U' : forall {Γ} -> Ty Γ
  U' = U
  _,'_ : forall Γ -> Ty Γ -> Con
  _,'_ = _,_

  data Sub : Con -> Con -> Set where
    coeˢ : forall {Γ Γ' Δ Δ'} -> Sub Γ Δ -> Γ ≡ˠ Γ' -> Δ ≡ˠ Δ' -> 
           Sub Γ' Δ'
    _•_  : forall {B Γ Δ} -> Sub Γ Δ -> Sub B Γ -> Sub B Δ
    id   : forall {Γ} -> Sub Γ Γ
    pop  : forall {Γ} σ -> Sub (Γ , σ) Γ
    _<_ : forall {Γ Δ σ}(ts : Sub Γ Δ) -> Tm Γ (σ [ ts ]⁺) -> 
          Sub Γ (Δ , σ)

  _•'_ : forall {B Γ Δ} -> Sub Γ Δ -> Sub B Γ -> Sub B Δ
  _•'_ = _•_
  pop'  : forall {Γ} σ -> Sub (Γ ,' σ) Γ
  pop' = pop


  data _≡⁺_ : forall {Γ Γ'} -> Ty Γ -> Ty Γ' -> Set where
    congU : forall {Γ}{Γ'} -> Γ ≡ˠ Γ' -> U {Γ} ≡⁺ U {Γ'}
    congEl : forall {Γ Γ'}{t : Tm Γ U}{t' : Tm Γ' U} -> Γ ≡ˠ Γ' -> t ≡ t' -> 
             El t ≡⁺ El t'
    congΠ : forall {Γ}{σ : Ty Γ}{τ : Ty (Γ , σ)} -> 
         forall {Γ'}{σ' : Ty Γ'}{τ' : Ty (Γ' , σ')} -> 
         (p : σ ≡⁺ σ')(q : τ ≡⁺ τ') ->
         Π σ τ ≡⁺ Π σ' τ'

  data _≡ˢ_ : forall {Γ Γ' Δ Δ'} -> Sub Γ Δ -> Sub Γ' Δ' -> Set where
    -- Setoid boilerplate
    cohˢ : forall {Γ Γ' Δ Δ'}(ts : Sub Γ Δ)(p : Γ ≡ˠ Γ')(q : Δ ≡ˠ Δ') ->
            coeˢ ts p q  ≡ˢ ts

    -- Equivalence boilerplate
    reflˢ : forall {Γ}{Δ}{ts : Sub Γ Δ} -> ts ≡ˢ ts
    transˢ : forall {Γ Γ' Γ''}{Δ Δ' Δ''} -> 
             {ts : Sub Γ Δ}{ts' : Sub Γ' Δ'}{ts'' : Sub Γ'' Δ''} -> 
             ts ≡ˢ ts' -> ts' ≡ˢ ts'' -> ts ≡ˢ ts''
    symˢ : forall {Γ Γ'}{Δ Δ'}{ts : Sub Γ Δ}{ts' : Sub Γ' Δ'} ->
           ts ≡ˢ ts' -> ts' ≡ˢ ts

    -- Subsitution boilerplate
    rightidˢ : forall {Γ}{Δ}{ts : Sub Γ Δ} -> (ts • id) ≡ˢ ts
    assocˢ : forall {A}{B}{Γ}{Δ}{ts : Sub Γ Δ}{us : Sub B Γ}{vs : Sub A B} ->
         (ts • us • vs) ≡ˢ ts • (us • vs)

    -- Congruence boilerplate
    cong• : forall {B B' Γ Γ' Δ Δ'}{ts : Sub Γ Δ}{us : Sub B Γ} ->
            {ts' : Sub Γ' Δ'}{us' : Sub B' Γ'} -> ts ≡ˢ ts' -> us ≡ˢ us' ->
            ts • us ≡ˢ ts' • us'
    congid : forall {Γ Γ'} -> Γ ≡ˠ Γ' -> id {Γ} ≡ˢ id {Γ'}
    congpop : forall {Γ Γ'}{σ : Ty Γ}{σ' : Ty Γ'} -> 
         σ ≡⁺ σ' ->  pop σ ≡ˢ pop σ'
    cong< : forall {Γ Γ'}{Δ Δ'}{σ}{σ'}{ts : Sub Γ Δ}{ts' : Sub Γ' Δ'}
          {t : Tm Γ (σ [ ts ]⁺)}{t' : Tm Γ' (σ' [ ts' ]⁺)} -> 
          ts ≡ˢ ts' -> t ≡ t' -> (ts < t) ≡ˢ (ts' < t')
    -- Computation rules
    leftid : forall {Γ}{Δ}{ts : Sub Γ Δ} -> (id • ts) ≡ˢ ts
    pop< : forall {Γ}{Δ}{σ}{ts : Sub Γ Δ}{t : Tm Γ (σ [ ts ]⁺)} -> 
           (pop σ • (ts < t)) ≡ˢ ts

    •< : forall {B}{Γ}{Δ}{σ} -> 
         {ts : Sub Γ Δ}{t : Tm Γ (σ [ ts ]⁺)}{us : Sub B Γ} -> 
         (ts < t) • us ≡ˢ (ts • us < coe' (t [ us ]') (assoc⁺ σ))
    poptop : forall {Γ}{σ} -> (pop σ < top') ≡ˢ id {Γ , σ}

  data Tm : forall Γ -> Ty Γ -> Set where
    coe  : forall {Γ Γ' σ σ'}-> Tm Γ σ -> σ ≡⁺ σ' -> Tm Γ' σ'
    _[_] : forall {Γ Δ σ} -> Tm Δ σ -> (ts : Sub Γ Δ) -> Tm Γ (σ [ ts ]⁺)
    top  : forall {Γ σ} -> Tm (Γ , σ) (σ [ pop σ ]⁺)
    λ    : forall {Γ σ τ} -> Tm (Γ , σ) τ -> Tm Γ (Π σ τ)
    app  : forall {Γ σ τ} -> Tm Γ (Π σ τ) -> Tm (Γ , σ) τ

  coe' : forall {Γ Γ' σ σ'}-> Tm Γ σ -> σ ≡⁺ σ' -> Tm Γ' σ'
  coe' = coe

  _[_]⁺ : forall {Γ Δ} -> Ty Δ -> Sub Γ Δ -> Ty Γ
  U        [ ts ]⁺ = U 
  El σ     [ ts ]⁺ = El (σ [ ts ]U) 
  Π σ τ    [ ts ]⁺ = Π (σ [ ts ]⁺) {! !} 
                       (τ [ ts  • pop _ < coe top (assoc⁺ σ) ]⁺)

  _[_]U : forall {Γ Δ} -> Tm Δ U' -> Sub Γ Δ -> Tm Γ U'
  t [ ts ]U = t [ ts ] 

  _[_]' : forall {Γ Δ σ} -> Tm Δ σ -> (ts : Sub Γ Δ) -> Tm Γ (σ [ ts ]⁺)
  _[_]' = _[_]


  top' : forall {Γ σ} -> Tm (Γ ,' σ) (σ [ pop' σ ]⁺)
  top' = top  

  data _≡_ : forall {Γ Γ' σ σ'} -> Tm Γ σ -> Tm Γ' σ' -> Set where
    -- Setoid boilerplate
    coh  : forall {Γ}{Γ'}{σ : Ty Γ}{σ' : Ty Γ'}(t : Tm Γ σ)(p : σ ≡⁺ σ') ->
           coe t p ≡ t

    -- Equivalence boilerplate
    refl : forall {Γ}{σ}{t : Tm Γ σ} -> t ≡ t
    sym : forall {Γ}{Γ'}{σ}{σ'}{t : Tm Γ σ}{t' : Tm Γ' σ'} -> 
          t ≡ t' -> t' ≡ t
    trans : forall {Γ Γ' Γ''}{σ}{σ'}{σ''} ->
           {t : Tm Γ σ}{t' : Tm Γ' σ'}{t'' : Tm Γ'' σ''} -> 
           t ≡ t' -> t' ≡ t'' -> t ≡ t''

    -- Substitution boilerplate
    rightid : forall {Γ}{σ : Ty Γ}{t : Tm Γ σ} -> t [ id ] ≡ t
    assoc : forall {B}{Γ}{Δ}{σ}{t : Tm Δ σ}{ts : Sub Γ Δ}{us : Sub B Γ} -> 
            t [ ts ] [ us ]  ≡  t [ ts • us ]
    cong : forall {Γ Γ'}{Δ Δ'}{σ : Ty Δ}{σ' : Ty Δ'} ->
          {t : Tm Δ σ}{t' : Tm Δ' σ'}{ts : Sub Γ Δ}{ts' : Sub Γ' Δ'} ->
          t ≡ t' -> ts ≡ˢ ts' ->
          t [ ts ] ≡ t' [ ts' ]

    -- Congruence boilerplate
    congtop : forall {Γ Γ'}{σ : Ty Γ}{σ' : Ty Γ'} -> σ ≡⁺ σ' -> 
              top {σ = σ} ≡ top {σ = σ'}
    congλ : forall {Γ Γ' σ σ' τ τ'}
         {t : Tm (Γ , σ) τ}{t' : Tm (Γ' , σ') τ'} -> σ ≡⁺ σ' ->
         t ≡ t' -> λ t ≡ λ t'
    congapp : forall {Γ Γ' σ σ' τ τ'} -> 
         {t : Tm Γ (Π σ τ)}{t' : Tm Γ' (Π σ' τ')} -> t ≡ t' -> app t ≡ app t'

    -- computation rules
    η : forall {Γ σ τ}{f : Tm Γ (Π σ τ)} -> 
        λ (app f) ≡ f
    β : forall {Γ σ τ}{t : Tm (Γ , σ) τ} ->
        app (λ t) ≡ t
    top< : forall {Γ Δ σ}{ts : Sub Γ Δ}{t : Tm Γ (σ [ ts ]⁺)} ->
           top {σ = σ} [ ts < t ] ≡ t
{-
    λ[] : forall {Γ Δ σ τ}{t : Tm (Δ , σ) τ}{ts : Sub Γ Δ} ->
          coe (λ t [ ts ]) Π[] ≡ λ (t [ ts ↗ σ ])
    app[] : forall {Γ Δ σ τ}{f : Tm Δ (Π σ τ)}{ts : Sub Γ Δ} ->
           app (coe (f [ ts ]) Π[]) ≡ app f [ ts ↗  σ ]

-}

  cong⁺ : forall {Γ Γ'}{Δ Δ'}{σ}{σ'}{ts : Sub Γ Δ}{ts' : Sub Γ' Δ'} -> 
          σ ≡⁺ σ' -> ts ≡ˢ ts' ->
          σ [ ts ]⁺ ≡⁺ σ' [ ts' ]⁺
  cong⁺ (congU p)    q = congU  (fogˢ q)
  cong⁺ (congEl p q) r = congEl (fogˢ r) (cong q r) 
  cong⁺ (congΠ p q)  r = 
    congΠ (cong⁺ p r)
          (cong⁺ q (cong< (cong• r (congpop (cong⁺ p r))) 
                          (trans (coh _ _) (trans (congtop (cong⁺ p r)) 
                                                  (sym (coh _ _))))))

  assoc⁺ : forall {B}{Γ}{Δ} σ{ts : Sub Γ Δ}{us : Sub B Γ} -> 
            σ [ ts ]⁺ [ us ]⁺  ≡⁺ σ [ ts •' us ]⁺
  assoc⁺ U     = congU (reflˠ _)
  assoc⁺ (El σ)  = congEl (reflˠ _)  assoc  
  assoc⁺ (Π σ τ) = 
    congΠ (assoc⁺ σ) 
          {! !}

  coe⁺  : forall {Γ Δ} -> Ty Γ -> Γ ≡ˠ Δ -> Ty Δ
  coe⁺ U       p = U
  coe⁺ (El σ)  p = El (coe σ (congU p))
  coe⁺ (Π σ τ) p = Π (coe⁺ σ p) (coe⁺ τ (cong, p (sym⁺ (coh⁺ σ p))))

  coh⁺ : forall {Γ}{Δ}(σ : Ty Γ)(p : Γ ≡ˠ Δ) -> coe⁺ σ  p ≡⁺ σ
  coh⁺ U       p = congU (symˠ p) 
  coh⁺ (El σ)  p = congEl (symˠ p) (coh σ  (congU p)) 
  coh⁺ (Π σ τ) p = congΠ (coh⁺ σ p) (coh⁺ τ _) 


  refl⁺ : forall {Γ}(σ : Ty Γ) -> σ ≡⁺ σ
  refl⁺ {Γ} U     = congU (reflˠ Γ)
  refl⁺ {Γ} (El σ)  = congEl (reflˠ Γ) refl  
  refl⁺ {Γ} (Π σ τ) = congΠ (refl⁺ σ) (refl⁺ τ)

  trans⁺ : forall {Γ}{Γ'}{Γ''}{σ : Ty Γ}{σ' : Ty Γ'}{σ'' : Ty Γ''} -> 
           σ ≡⁺ σ' -> σ' ≡⁺ σ'' -> σ ≡⁺ σ''
  trans⁺ (congU p)    (congU p')     = congU (transˠ p p') 
  trans⁺ (congEl p q) (congEl p' q') = congEl (transˠ p p') (trans q q') 
  trans⁺ (congΠ p q)  (congΠ p' q')  = congΠ (trans⁺ p p') (trans⁺ q q') 

  sym⁺ : forall {Γ}{Γ'}{σ : Ty Γ}{σ' : Ty Γ'} -> σ ≡⁺ σ' -> σ' ≡⁺ σ
  sym⁺ (congU p)    = congU (symˠ p) 
  sym⁺ (congEl p q) = congEl (symˠ p) (sym q) 
  sym⁺ (congΠ p q)  = congΠ (sym⁺ p) [ ts ](sym⁺ q) 

  reflˠ : forall Γ -> Γ ≡ˠ Γ
  reflˠ ε = congε 
  reflˠ (Γ , σ) = cong, (reflˠ Γ) (refl⁺ σ)

  symˠ : forall {Γ Γ'} -> Γ ≡ˠ Γ' -> Γ' ≡ˠ Γ
  symˠ congε  = congε 
  symˠ (cong, p q)  = cong, (symˠ p) (sym⁺ q) 

  transˠ : forall {Γ Γ' Γ''} -> Γ ≡ˠ Γ' -> Γ' ≡ˠ Γ'' -> Γ ≡ˠ Γ''
  transˠ congε congε = congε 
  transˠ (cong, p q) (cong, p' q') = cong, (transˠ p p') (trans⁺ q q') 

  fog⁺ : forall {Γ Γ'}{σ : Ty Γ}{σ' : Ty Γ'} -> σ ≡⁺ σ' -> Γ ≡ˠ Γ'
  fog⁺ (congU p) = p 
  fog⁺ (congEl p q) = p
  fog⁺ (congΠ p q) = fog⁺ p

  fog : forall {Γ Γ' σ σ'}{t : Tm Γ σ}{t' : Tm Γ' σ'} -> t ≡ t' ->
                σ ≡⁺ σ'
  fog (coh t p) = sym⁺ p
  fog refl = refl⁺ _
  fog (trans p q) = trans⁺ (fog p) (fog q)
  fog (sym p) = sym⁺ (fog p)
  fog rightid = {! !}
  fog assoc = assoc⁺ _
  fog (congtop p) = cong⁺ p (congpop p) 
  fog (cong p q) = cong⁺ (fog p) q
  fog (congλ p q) = congΠ p (fog q) 
  fog (congapp p) with fog p
  ... | congΠ p' q = q 
  fog β = refl⁺ _
  fog η = refl⁺ _
  fog top< = trans⁺ (assoc⁺ _) (cong⁺ (refl⁺ _) pop<)

  fogˢ : forall {Γ Γ' Δ Δ'}{ts : Sub Γ Δ}{ts' : Sub Γ' Δ'} -> 
                   ts ≡ˢ ts' -> Γ ≡ˠ Γ'
  fogˢ (cohˢ ts p q) = symˠ p
  fogˢ reflˢ = reflˠ _
  fogˢ (transˢ p q) = transˠ (fogˢ p) (fogˢ q)
  fogˢ (symˢ p) = symˠ (fogˢ p)
  fogˢ rightidˢ = reflˠ _
  fogˢ assocˢ = reflˠ _ 
  fogˢ (cong• p q) = fogˢ q
  fogˢ (congid p) = p
  fogˢ (congpop p) = cong, (fog⁺ p) p
  fogˢ (cong< p q) = fogˢ p
  fogˢ leftid = reflˠ _
  fogˢ pop< = reflˠ _
  fogˢ •< = reflˠ _
  fogˢ poptop = reflˠ _
